<?php
/**
 * Typography Helper
 *
 * PHP Version 5.3.1とCakePHP 1.3.11で動作を確認。
 *
 * @author http://purl.org/meta/me/
 * @version Typography Helper 1.0.1 2011-09-17
 * @license Creative Commons License CC BY (http://creativecommons.org/licenses/by/3.0/)
 */

App::import('Helper', 'Text');

App::import('Core', 'String');

/**
 * TypographyHelperクラスはタイポグラフィを手助けするヘルパーです。
 *
 * <code>
 * class ExamplesController extends AppController {
 *     var $helpers = array('Typography');
 * }
 * </code>
 */
class TypographyHelper extends TextHelper {

	/**
	 * List of helpers used by this helper
	 *
	 * @var array
	 */
	var $helpers = array('Html');

	/**
	 * 特別な文字の変換リスト
	 *
	 * @var array
	 * @access public
	 */
	var $specialCharacters = array(
		'(c)' => '&#169;',
		'(C)' => '&#169;',
		'(r)' => '&#174;',
		'(R)' => '&#174;',
		'(tm)' => '&#8482;',
		'(TM)' => '&#8482;',
		'(sm)' => '&#8480;',
		'(SM)' => '&#8480;',
		'(p)' => '&#8471;',
		'(P)' => '&#8471;',
		'...' => '&#8230;'
	);

	/**
	 * 数学に関する記号の変換リスト
	 *
	 * @var array
	 * @access public
	 */
	var $mathCharacters = array(
		'/([0-9.]+) ?x ?([0-9.]+)/' => '$1&#215;$2',
		'/([0-9.]+) ?X ?([0-9.]+)/' => '$1&#215;$2',
		'/([0-9.]+) ?\* ?([0-9.]+)/' => '$1&#215;$2',
		'/([0-9]+)\/([0-9]+)/' => '<sup>$1</sup>&#8260;<sub>$2</sub>',
		'/([0-9.]+)\^([0-9]+)/' => '$1<sup>$2</sup>'
	);

	/**
	 * Default map of accented and special characters to ASCII characters
	 *
	 * @var array
	 * @access public
	 * @link http://api.cakephp.org/class/inflector
	 */
	var $transliteration = array(
		'/ä|æ|ǽ/' => 'ae',
		'/ö|œ/' => 'oe',
		'/ü/' => 'ue',
		'/Ä/' => 'Ae',
		'/Ü/' => 'Ue',
		'/Ö/' => 'Oe',
		'/À|Á|Â|Ã|Ä|Å|Ǻ|Ā|Ă|Ą|Ǎ/' => 'A',
		'/à|á|â|ã|å|ǻ|ā|ă|ą|ǎ|ª/' => 'a',
		'/Ç|Ć|Ĉ|Ċ|Č/' => 'C',
		'/ç|ć|ĉ|ċ|č/' => 'c',
		'/Ð|Ď|Đ/' => 'D',
		'/ð|ď|đ/' => 'd',
		'/È|É|Ê|Ë|Ē|Ĕ|Ė|Ę|Ě/' => 'E',
		'/è|é|ê|ë|ē|ĕ|ė|ę|ě/' => 'e',
		'/Ĝ|Ğ|Ġ|Ģ/' => 'G',
		'/ĝ|ğ|ġ|ģ/' => 'g',
		'/Ĥ|Ħ/' => 'H',
		'/ĥ|ħ/' => 'h',
		'/Ì|Í|Î|Ï|Ĩ|Ī|Ĭ|Ǐ|Į|İ/' => 'I',
		'/ì|í|î|ï|ĩ|ī|ĭ|ǐ|į|ı/' => 'i',
		'/Ĵ/' => 'J',
		'/ĵ/' => 'j',
		'/Ķ/' => 'K',
		'/ķ/' => 'k',
		'/Ĺ|Ļ|Ľ|Ŀ|Ł/' => 'L',
		'/ĺ|ļ|ľ|ŀ|ł/' => 'l',
		'/Ñ|Ń|Ņ|Ň/' => 'N',
		'/ñ|ń|ņ|ň|ŉ/' => 'n',
		'/Ò|Ó|Ô|Õ|Ō|Ŏ|Ǒ|Ő|Ơ|Ø|Ǿ/' => 'O',
		'/ò|ó|ô|õ|ō|ŏ|ǒ|ő|ơ|ø|ǿ|º/' => 'o',
		'/Ŕ|Ŗ|Ř/' => 'R',
		'/ŕ|ŗ|ř/' => 'r',
		'/Ś|Ŝ|Ş|Š/' => 'S',
		'/ś|ŝ|ş|š|ſ/' => 's',
		'/Ţ|Ť|Ŧ/' => 'T',
		'/ţ|ť|ŧ/' => 't',
		'/Ù|Ú|Û|Ũ|Ū|Ŭ|Ů|Ű|Ų|Ư|Ǔ|Ǖ|Ǘ|Ǚ|Ǜ/' => 'U',
		'/ù|ú|û|ũ|ū|ŭ|ů|ű|ų|ư|ǔ|ǖ|ǘ|ǚ|ǜ/' => 'u',
		'/Ý|Ÿ|Ŷ/' => 'Y',
		'/ý|ÿ|ŷ/' => 'y',
		'/Ŵ/' => 'W',
		'/ŵ/' => 'w',
		'/Ź|Ż|Ž/' => 'Z',
		'/ź|ż|ž/' => 'z',
		'/Æ|Ǽ/' => 'AE',
		'/ß/'=> 'ss',
		'/Ĳ/' => 'IJ',
		'/ĳ/' => 'ij',
		'/Œ/' => 'OE',
		'/ƒ/' => 'f'
	);

	/**
	 * 単位に関する記号の変換リスト
	 *
	 * @var array
	 * @access public
	 */
	var $unitCharacters = array(
		'/([0-9.]+) ?°C/' => '$1&#8451;',
		'/([0-9.]+) ?°F/' => '$1&#8457;',
		'/No. ([0-9]+)/' => '&#8470; $1',
		'/([0-9]+) ?(L|l) /' => '$1 &#8467; ',
		'/([0-9]+) ?(µl|µL) /' => '$1 &#13205; ',
		'/([0-9]+) ?(ml|mL|ML) /' => '$1 &#13206; ',
		'/([0-9]+) ?(dl|dL|DL) /' => '$1 &#13207; ',
		'/([0-9]+) ?(kl|kL|KL) /' => '$1 &#13208; ',
	);

	/**
	 * UUID
	 *
	 * @var string
	 * @access protected
	 */
	var $_uuid = null;

	/**
	 * Before render callback. beforeRender is called before the view file is rendered.
	 *
	 * Overridden in subclasses.
	 *
	 * @return void
	 * @access public
	 */
	function beforeRender() {
		parent::beforeRender();

		$this->_uuid = String::uuid();
	}

	/**
	 * n回連続したダッシュを長いダッシュに変換します。
	 *
	 * <code>
	 * $typography->horizontal('-------');
	 *
	 * # 出力すると以下のようになります。
	 * # ————
	 * </code>
	 *
	 * @param string $string 処理を行う文字列
	 * @param integer $int 何回連続したダッシュを変換するのか
	 * @return string n回連続したダッシュが長いダッシュに変換された文字列
	 * @access public
	 */
	function horizontal($string, $int = 2) {
		return preg_replace_callback(
			'/-{' . $int . ',}/',
			create_function(
				'$matches',
				'return strtr($matches[0], array("--" => "—", "-" => "—"));'
			),
			$string
		);
	}

	/**
	 * 数学記号に変換します。
	 *
	 * <code>
	 * $typography->mathCharacters('Multiplication 100x100 Fraction 1/2');
	 *
	 * # 出力すると以下のようになります。
	 * # Multiplication 100&#215;100 Fraction <sup>1</sup>&#8260;<sub>2</sub>
	 * </code>
	 *
	 * @param string $string 処理を行う文字列
	 * @return string 数学記号に変換された文字列
	 * @access public
	 */
	function mathCharacters($string) {
		return preg_replace(array_keys($this->mathCharacters), array_values($this->mathCharacters), $string);
	}

	/**
	 * 基本的にはnl2brと同じ処理をしますが特定の要素内での処理は除くことができます。
	 *
	 * <code>
	 * $typography->nl2brExcept("foo isn't\n bar\n\n<pre>\n pre Element\n</pre>");
	 *
	 * # 出力すると以下のようになります。
	 * # foo isn't<br />
	 * #  bar<br />
	 * # <br />
	 * # <pre>
	 * #  pre Element
	 * # </pre>
	 * </code>
	 *
	 * @param string $string 入力する文字列
	 * @param array $except 処理を除く要素の配列
	 * @param boolean $is_xhtml trueの場合はXHTMLに対応した&lt;br /&gt;、falseの場合はHTMLに対応した&lt;br&gt;になります。
	 * @return string 改行がbr要素に変換された文字列
	 * @access public
	 */
	function nl2brExcept($string, $except = array('pre'), $is_xhtml = true) {
		$newstr = $this->__replaceEncode($string, $except);

		if(version_compare(PHP_VERSION, '5.3.0') >= 0) {
			$newstr = nl2br($newstr, $is_xhtml);
		} else {
			$newstr = nl2br($newstr);
		}

		return $this->__replaceDecode($newstr);
	}

	/**
	 * 改行の前にリターン記号を追加します。
	 *
	 * <code>
	 * $typography->nl2carriageReturn("foo\nbar");
	 *
	 * # 出力すると以下のようになります。
	 * # foo↵
	 * # bar
	 * </code>
	 *
	 * @param string $string 処理を行う文字列
	 * @param return $return リターン記号
	 * @return string 変換された文字列
	 * @access public
	 */
	function nl2carriageReturn($string, $return = '↵') {
		return strtr($string, array("\r\n" => $return . "\r\n", "\r" => $return . "\r", "\n" => $return . "\n"));
	}

	/**
	 * URLを正規化します。
	 *
	 * <code>
	 * $typography->normalizeUrl('hxxp URL ttp://example.com/ or hxxp://example.com/example/');
	 *
	 * # 出力すると以下のようになります。
	 * # hxxp URL http://example.com/ or http://example.com/example/
	 * </code>
	 *
	 * @param string $string 処理を行う文字列
	 * @return string URLを正規化した文字列
	 * @access public
	 */
	function normalizeUrl($string) {
		$noPchar = '[-_.!~*\'()a-zA-Z0-9;\/?:\@&=+\$,%#]+';
		return preg_replace('/(htp|ttp|tp|hxxp|h\+\+p|h\*\*p||ht_tp|htt|_ttp)(s?:\/\/' . $noPchar . ')/', 'http$2', $string);
	}

	/**
	 * URLを正規化してリンクに変換します。
	 *
	 * <code>
	 * $typography->normalizeUrlLink('hxxp URL ttp://example.com/ or hxxp://example.com/example/', array('rel' => 'nofollow'));
	 *
	 * # 出力すると以下のようになります。
	 * # hxxp URL <a href="http://example.com/" rel="nofollow">http://example.com/</a> or <a href="http://example.com/example/" rel="nofollow">http://example.com/example/</a>
	 * </code>
	 *
	 * @param string $string 処理を行う文字列
	 * @param array $options Array of HTML options.
	 * @return string URLを正規化してリンクに変換された文字列
	 * @access public
	 */
	function normalizeUrlLink($string, $options = array()) {
		return $this->autoLink($this->normalizeUrl($string), $options);
	}

	/**
	 * a要素内のURLを短くします。
	 *
	 * <code>
	 * echo $typography->shortenUrl('The Aristotle <a href="http://ja.wikipedia.org/wiki/%E3%82%A2%E3%83%AA%E3%82%B9%E3%83%88%E3%83%86%E3%83%AC%E3%82%B9">http://ja.wikipedia.org/wiki/%E3%82%A2%E3%83%AA%E3%82%B9%E3%83%88%E3%83%86%E3%83%AC%E3%82%B9</a>', 30, '…');
	 *
	 * # 出力すると以下のようになります。
	 * # The Aristotle <a href="http://ja.wikipedia.org/wiki/%E3%82%A2%E3%83%AA%E3%82%B9%E3%83%88%E3%83%86%E3%83%AC%E3%82%B9">http://ja.wikipedia.org/wiki/%…</a>
	 * </code>
	 *
	 * @param string $string 入力する文字列
	 * @param integer $length 何文字まで切り詰めるか
	 * @param string $ending 切り詰めた後の文字列
	 * @return string a要素内のURLを短くされた文字列
	 * @access public
	 */
	function shortenUrl($string, $length = 50, $ending = null) {
		$ending = var_export($ending, true);

		return preg_replace_callback(
			'/(<a.*?>)((https?|ftp)(:\/\/[-_.!~*\'()a-zA-Z0-9;\/?:\@&=+\$,%#]+?))<\/a>/',
			create_function(
				'$matches',
				'return $matches[1] . substr($matches[2], 0, ' . $length . ') . ' . $ending . ' . "</a>";'
			),
			$string
		);
	}

	/**
	 * 一部の記述を記号文字へ変換します。
	 *
	 * <code>
	 * $typography->specialCharacters('Copyright (C) 2011 example.com All Rights Reserved.');
	 *
	 * # 出力すると以下のようになります。
	 * # Copyright &#169; 2011 example.com All Rights Reserved.
	 * </code>
	 *
	 * @param string $string 処理を行う文字列
	 * @return string 変換された文字列
	 * @access public
	 */
	function specialCharacters($string) {
		return strtr($string, $this->specialCharacters);
	}

	/**
	 * 改行やタブ文字などを消去します。
	 *
	 * <code>
	 * $typography->trim("<p>\r\nWelcome This is my HTML document\r\n</p>");
	 *
	 * # 出力すると以下のようになります。
	 * # <p>Welcome This is my HTML document</p>
	 * </code>
	 *
	 * @param string $string 入力する文字列
	 * @return string 改行やタブ文字などを消去された文字列
	 * @access public
	 */
	function trim($string) {
		return str_replace(array("\r", "\n", "\t", "\0", "\x0B"), '', $string);
	}

	/**
	 * 基本的にはtrimと同じ処理をしますが特定の要素内での処理は除くことができます。
	 *
	 * <code>
	 * $typography->trimExcept("<p>\r\nWelcome This is my HTML document\r\n</p>\r\n<pre>\r\n pre Element\r\n</pre>");
	 *
	 * # 出力すると以下のようになります。
	 * # <p>Welcome This is my HTML document</p><pre>
	 * #  pre Element
	 * # </pre>
	 * </code>
	 *
	 * @param string $string 入力する文字列
	 * @param array $except 処理を除く要素の配列
	 * @return string 改行やタブ文字などを消去された文字列
	 * @access public
	 */
	function trimExcept($string, $except = array('pre', 'script')) {
		$newstr = $this->__replaceEncode($string, $except);

		$newstr = $this->trim($newstr);

		return $this->__replaceDecode($newstr);
	}

	/**
	 * 2回以上連続するスペースとタブ文字を1つのスペースに変換します。
	 *
	 * <code>
	 * $typography->trimScrapes("Double Scrape  and\tTab");
	 *
	 * # 出力すると以下のようになります。
	 * # Double Scrape and Tab
	 * </code>
	 *
	 * @param string $string 処理を行う文字列
	 * @return string 変換された文字列
	 * @access public
	 */
	function trimScrapes($string) {
		return preg_replace(array('/ {2,}/', '/\\t{1,}/'), ' ', $string);
	}

	/**
	 * 文字列を一行にし、2回以上連続するスペースを1つにします。
	 *
	 * <code>
	 * $typography->oneLine("oneLine Test.\r\nDouble Scrape  and\tTab");
	 *
	 * # 出力すると以下のようになります。
	 * # oneLine Test. Double Scrape and Tab
	 * </code>
	 *
	 * @param string $string 処理を行う文字列
	 * @return string 一行に変換された文字列
	 * @access public
	 */
	function oneLine($string) {
		$string = str_replace(array("\r", "\n", "\0", "\x0B"), ' ', $string);
		return $this->trimScrapes($string);
	}

	/**
	 * 改行をp要素とbr要素に変換します。
	 *
	 * 2回以上連続する改行をp要素に変換します。単一の改行はbr要素に変換します。
	 *
	 * <code>
	 * $typography->paragraph("paragraph method Test\r\nbr Elements\r\n\r\np Elements\r\n\r\n\r\nend.");
	 *
	 * # 出力すると以下のようになります。
	 * # <p>paragraph method Test<br />
	 * # br Elements</p><p>p Elements</p><p>end.</p>
	 * </code>
	 *
	 * @param string $string 処理を行う文字列
	 * @param array $options p要素のオプション
	 * @return string 改行がp要素またはbr要素に変換された文字列
	 * @access public
	 */
	function paragraph($string, $options = array()) {
		if(isset($options['is_xhtml'])) {
			$is_xhtml = $options['is_xhtml'];

			unset($options['is_xhtml']);
		} else {
			$is_xhtml = true;
		}

		$string = str_replace("\r", '', $string);

		$array = explode("\n\n", $string);

		$out = '';
		foreach($array as $key => $value) {
			if(trim($value)) {
				$pElement = $this->Html->para(null, trim($value), $options);

				if(version_compare(PHP_VERSION, '5.3.0') >= 0) {
					$out .= nl2br($pElement, $is_xhtml);
				} else {
					$out .= nl2br($pElement);
				}
			}
		}

		return $out;
	}

	/**
	 * タブ文字をスペースに変換します。
	 *
	 * <code>
	 * echo $typography->tabToScrape("tabToScrape\tTest");
	 *
	 * # 出力すると以下のようになります。
	 * # tabToScrape    Test
	 * </code>
	 *
	 * @param string $string 処理を行う文字列
	 * @param integer $multiplier タブ文字をスペース何個分に変換するか
	 * @return string タブ文字がスペースに変換された文字列
	 * @access public
	 */
	function tabToScrape($string, $multiplier = 4) {
		return strtr($string, array("\t" => str_repeat(' ', $multiplier)));
	}

	/**
	 * ASCI文字に変換します。
	 *
	 * @param string $string 処理を行う文字列
	 * @return string ASCI文字に変換された文字列
	 * @access public
	 */
	function transliteration($string) {
		return preg_replace(array_keys($this->transliteration), array_values($this->transliteration), $string);
	}

	/**
	 * 単位記号に変換します。
	 *
	 * <code>
	 * $typography->unitCharacters('100 L 100 °C 100 °C No. 100');
	 *
	 * # 出力すると以下のようになります。
	 * # 100 &#8467; 100&#8451; 100&#8451; &#8470; 100
	 * </code>
	 *
	 * @param string $string 処理を行う文字列
	 * @return string 単位記号に変換された文字列
	 * @access public
	 */
	function unitCharacters($string) {
		return preg_replace(array_keys($this->unitCharacters), array_values($this->unitCharacters), $string);
	}

	function __replaceEncode($string, $except = array()) {
		if(is_string($except)) {
			$except = array($except);
		}

		$replace = '/' . implode('|', array_map(
			create_function(
				'$value',
				'
					if(is_array($value) && isset($value["start"]) && isset($value["end"])) {
						return preg_quote($value["start"], "/") . ".*?" . preg_quote($value["end"], "/");
					}
					return "<" . $value . ".*?>.*?<\\/" . $value . ">";
				'
			),
			$except
		)) . '/s';

		return preg_replace_callback($replace, create_function('$matches', '
			return "<" . base64_encode($matches[0]) . "<' . $this->_uuid . '>>";
		'), $string);
	}

	function __replaceDecode($string) {
		return preg_replace_callback('/<([\w+\/=]+?)<' . $this->_uuid . '>>/', create_function('$matches', '
			return base64_decode($matches[1]);
		'), $string);
	}
}
?>